/*
This class and file handle the basic functionality of the connection state for MetaMask.

Note that this class is created before the DOM functionality is available to reference.

The states

if MetaMask is not installed then button reads "Learn how to install MetaMask!" and onClick launches MetaMask Website.
As long as it's not installed, the user can press the button as many times as they would like.

If MetaMask is installed then the button reads "Connect MetaMask" and onClick it calls eth_requestAccounts to link
MetaMask to the website. 

The next change of status happens in the onClick handler. Either MetaMask will return success - which instructs the
code to change the button text to "0x123...4567 connected" or failure which changes the button back to "Connect MetaMask"
which is the state that leads to calling eth_requestAccounts.

Note that if MetaMask is already connected, the button is disabled, it shows the linked wallet.

eth_requestAccounts -> This launches MetaMask so the user can link their account. This call should only be
    made if MetaMask is not connected. It is also asynchronous and waits for the user to finish the connection
    process. When it's done and returns, the return value is an array of accounts that have been connected.
    If the request fails, there are no linked addresses - and we know that the user intentionally did not
    link the account.

eth_accounts -> This gets the linked account number. If MetaMask is not linked, it comes back empty (Array 0).
    Also, if MetaMask is connected, this call gets the linked wallet address without launching the MetaMask
    user experience. Thus, it's a private way of getting the linked wallet address. Because it's an RPC call
    it blocks (which is only a fraction of a second)

This class requires a button with id="connectButtonId" on the page. All other HTML items are optional.
*/
import * as dom from './domutils.js'

const thetaNetId = 361

// These are the names of the three items we're going to look for on the page.
// Each item will be defened based on it's existance. 
const chainIdStr = "chainId"
const accountIdStr = "accountId"
const connectButtonStr = "connectButtonId"
const statusIdStr = "cbStatusId"
const balanceIdStr = "balanceId"
const blockIdStr = "blockId"


const isMetaMaskInstalled = () => {
    const { ethereum } = window
    return Boolean(ethereum && ethereum.isMetaMask)
}

class cconnectbutton {

    #chainIdDiv
    #chainVal  // Required

    #accountIdDiv
    #accounts  // required

    #connectButton // required
    #statusIdDiv

    #balanceIdDiv
    #balanceValue

    #blockIdDiv
    #blockValue

    #installed

    #turnOnDebug

    constructor() {
        this.#chainVal = 0
        this.#accounts = null
        this.#installed = false

        // Can't access any element on the page.
        this.#chainIdDiv = null

        this.#accountIdDiv = null

        this.#connectButton = null
        this.#statusIdDiv = null

        this.#balanceIdDiv = null
        this.#balanceValue = 0

        this.#blockIdDiv = null
        this.#blockValue = 0

        this.#turnOnDebug = false;
        this.#DebugSpew("cconnectButton constructor")
    }

    turnOnDebug(b=false) {
        this.#turnOnDebug = b
    }

    #DebugSpew(Str) {
        if (true == this.#turnOnDebug) {
            console.log(Str)
        }
    }

    #initHTML() {

        this.#DebugSpew("cb.#initHTML() enter")

        this.getButton()
        if (null == this.#chainIdDiv ) {
            this.#chainIdDiv = dom.fetch(this.#chainIdDiv, chainIdStr)
        }
        if (null == this.#accountIdDiv ) {
            this.#accountIdDiv = dom.fetch(this.#accountIdDiv, accountIdStr)
        }
        if (null == this.#balanceIdDiv ) {
            this.#balanceIdDiv = dom.fetch(this.#balanceIdDiv, balanceIdStr)
            this.setBalance(0)
        }
        if (null == this.#blockIdDiv ) {
            this.#blockIdDiv = dom.fetch(this.#blockIdDiv, blockIdStr)
            this.setBlock(0)
        }
        if (null == this.#statusIdDiv ) {
            this.#statusIdDiv = dom.fetch(this.#statusIdDiv, statusIdStr)
        }
        this.#DebugSpew("cb.#initHTML() leave")
    }


    // fetches the button and returns true or false based on finding it.
    exists() {
        if (this.getButton() != null) {
            this.#DebugSpew("cb.exists() returning true")
        }
        return (this.getButton() != null)
    }
    getButton() {
        if (null == this.#connectButton) {
            this.#connectButton = dom.fetch(this.#connectButton, connectButtonStr)
        }
        return this.#connectButton
    }

    // This will return true when the class has an account on the Theta Network
    ready() {
        if (this.isThetaNet() && null != this.getAccount()) {
            this.#DebugSpew("cb.ready() returning true")
            return true;
        }
        this.#DebugSpew("cb.ready() returning false")
        return false
    }

    isMetaMaskInstalled() {
        if (isMetaMaskInstalled()) {
            this.#DebugSpew("cb.isMetaMaskInstalled() returning true")
            this.#installed = true;
            return true
        }
        this.#installed = false;
        this.#DebugSpew("cb.isMetaMaskInstalled() returning false")
        return false
    }


    // This could be called at any time
    handleChainIdChange(chainId) {
        this.#initHTML()

        let newChainVal = parseInt(chainId, 16)
        if (this.#chainVal != newChainVal) {
            this.#chainVal = newChainVal
            // the chainId changed! all the data we've gathered becomes invalid
            this.#DebugSpew("cb.handleChainIdChange() changing to " + this.#chainVal)

        }
        if (dom.exists(this.#chainIdDiv, chainIdStr)) {
            this.#chainIdDiv.innerHTML = this.chainName()
        }
        this.#updateStatus()
    }

    chainId() {
        return this.#chainVal
    }
    isThetaNet() {
        return (thetaNetId == this.#chainVal)
    }
    chainName() {
        if (this.isThetaNet() ) {
            return "Theta Network"
        }
        else {
            return ("Non Theta Blockchain id: " +(this.#chainVal).toString(10) )
        }
    }

    // can get called at any time.
    handleAccountChange(account) {
        this.#initHTML()

        this.#DebugSpew("cb.handleAccountChange() enter")
        // did it change?
        let oldAccount = this.getAccount()

        // now check if we have something useful
        this.#accounts = null
        if (null != account) {
            if (account.length > 0) {
                // have account
                this.#accounts = account
            }
        }

        if (null != this.#accounts ) {
            // is it changing?
            if (oldAccount != this.#accounts ) {
                // account changed!
                this.#DebugSpew("cb.handleAccountChange() changing to: " + this.getAccount())

                // set the element
                if (this.accountExists()) {
                    if (dom.exists(this.#accountIdDiv, accountIdStr)) {
                        this.#accountIdDiv.innerHTML = this.getAccount()
                    }
                }
            }
            else {
                // account did not change!
                this.#DebugSpew("cb.handleAccountChange() did not change")
            }
        }
        else {
            this.#DebugSpew("cb.handleAccountChange() account is null")
        }

        this.#DebugSpew("cb.handleAccountChange() leave")
    }
    getAccount(Index = 0) {
        if (null == this.#accounts ) {
            return null
        }
        return this.#accounts[0]
    }

    accountExists() {
        this.#DebugSpew("cb.accountExists()")
        if (this.#accounts && this.#accounts.length > 0) {
            return true
        }
        return false
    }
/*
    accountsNumber() {
        if (this.#accounts && this.#accounts.length > 0) {
            return this.#accounts.length
        }
        this.#DebugSpew("called accountsNumber() and there is no accounts to check length of")
        return 0
    }
*/

    #shortAccounts(index) {
        let addr = this.#accounts[index]
        if (addr.length == 42) {
            let str = addr[0] + addr[1] + addr[2] + addr[3] + addr[4] + "..." + addr[38] + addr[39] + addr[40] + addr[41]
            return str
        }
        else {
            return "Invalid Address Length"
        }
    }

    #updateStatus() {
        if (null != this.#statusIdDiv ) {
            // Now update the status
            if (this.isThetaNet()) {
                this.#statusIdDiv.style.display = "none"
                //this.#statusIdDiv.innerText = "Correct: Theta Network"
            }
            else {
                this.#statusIdDiv.style.display = ""
                this.#statusIdDiv.innerText = "Connected to Unknown Network(" + this.chainId() + ")"
            }
        }
    }

    // This routine sets all the html elements to their proper settings
    updateState() {
        this.#DebugSpew("cb.updateState() enter")

        if (this.#installed ) {
            // MetaMask is installed
            // Either we're connected or not connected. If we're connected we'll have
            // an account. but could have a chainId that is anything.
            this.#DebugSpew("cb.updateState() MetaMask installed")

            if (null != this.getAccount()) {
                // we're connected
                this.#DebugSpew("cb.updateState() we're connected")

                // setup the state of the button
                this.#connectButton.innerText = this.#shortAccounts(0) + " connected"
                this.#connectButton.disabled = true // button not clickable

            }
            else {
                // we're not connected
                this.#DebugSpew("cb.updateState() we're NOT connected")

                this.#connectButton.innerText = 'Connect MetaMask'
                this.#connectButton.disabled = false // button is clickable

                // Need to clear out Account element.
                if (null != this.#accountIdDiv ) {
                    this.#accountIdDiv.innerText = ""
                }
            }
        }
        else {
            // Metamask is not installed.
            this.#DebugSpew("cb.updateState() should read install MetaMask")

            this.#connectButton.innerText = 'Learn how to install MetaMask!'
            this.#connectButton.disabled = false

            // clear out all the other items
        }

        if (!this.ready()) {
            // make sure all the extra elements are
            if (null != this.#balanceIdDiv ) {
                this.#balanceIdDiv.innerText = ""
            }
            if (null != this.#blockIdDiv ) {
                this.#blockIdDiv.innerText = ""
            }
        }

        this.#DebugSpew("cb.updateState() leave")
    }
    isInstall() {
         return !this.#installed
    }
    isConnect() {
        return this.#installed
    }


    // Extra functionality here...

    setBalance(bal) {
        this.#balanceValue = bal
        if (dom.exists(this.#balanceIdDiv, balanceIdStr)) {
            this.#balanceIdDiv.innerHTML = this.balanceTag()
        }
    }
    balance() {
        return this.#balanceValue
    }
    balanceTag() {
        return (this.#balanceValue + " Tfuel")
    }

    setBlock(val) {
        let dec = parseInt(val, 16)
        this.#blockValue = dec
        if (dom.exists(this.#blockIdDiv, blockIdStr)) {
            let n = this.#blockValue
            this.#blockIdDiv.innerHTML = String(n).replace(/(.)(?=(\d{3})+$)/g, '$1,')
        }
    }
    block() {
        return this.#blockValue
    }

}

export { cconnectbutton }